@@ -215,12 +215,12 @@ class Protocol:
     def wait_for_io_threads_exit(self, timeout=None):
         io_threads = [x for x in (self._read_thread, self._write_thread) if x is not None]
         for t in io_threads:
-            if t.isAlive():
+            if t.is_alive():
                 t.join(timeout)
         exited = True
         cinfo = self._conn or "cleared connection"
         for t in io_threads:
-            if t.isAlive():
+            if t.is_alive():
                 log.warn("Warning: %s thread of %s is still alive (timeout=%s)", t.name, cinfo, timeout)
                 exited = False
         return exited
@@ -494,12 +494,15 @@ class Protocol:
             self.enable_compressor("none")
             return
         opts = compression.get_enabled_compressors(order=compression.PERFORMANCE_ORDER)
+        compressors = caps.strtupleget("compressors")
         log("enable_compressor_from_caps(..) options=%s", opts)
         for c in opts:      #ie: [zlib, lz4, lzo]
-            if caps.boolget(c):
+            if c=="none":
+                continue
+            if c in compressors or caps.boolget(c):
                 self.enable_compressor(c)
                 return
-        log.warn("compression disabled: no matching compressor found")
+        log.warn("Warning: compression disabled, no matching compressor found")
         self.enable_compressor("none")
 
     def enable_compressor(self, compressor):
