@@ -146,6 +146,9 @@ def do_get_socket_dirs():
         SOCKET_DIRS.append("/run/xpra")
     elif os.path.exists("/var/run"):
         SOCKET_DIRS.append("/var/run/xpra")
+    #Debian and Ubuntu often don't create a reliable XDG_RUNTIME_DIR
+    #other distros may not create one when using "su"
+    SOCKET_DIRS.append("~/.xpra")
     return SOCKET_DIRS
 
 def do_get_default_log_dirs():
