@@ -1,5 +1,5 @@
 # This file is part of Xpra.
-# Copyright (C) 2011-2018 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2011-2020 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -7,44 +7,53 @@ import os
 import re
 
 #ensure that we use gtk as display source:
-from xpra.x11.gtk_x11.gdk_display_source import init_display_source
-init_display_source()
-
-from xpra.util import std, csv
-from xpra.gtk_common.error import xsync
+from xpra.x11.gtk_x11.gdk_display_source import init_gdk_display_source
+from xpra.util import std, csv, envbool
+from xpra.os_util import bytestostr
+from xpra.gtk_common.error import xsync, xlog
 from xpra.x11.bindings.keyboard_bindings import X11KeyboardBindings #@UnresolvedImport
+from xpra.log import Logger
+
+init_gdk_display_source()
 X11Keyboard = X11KeyboardBindings()
 
-from xpra.log import Logger
 log = Logger("x11", "keyboard")
 
+XKB = envbool("XPRA_XKB", True)
 DEBUG_KEYSYMS = [x for x in os.environ.get("XPRA_DEBUG_KEYSYMS", "").split(",") if len(x)>0]
 
 #keys we choose not to map if the free space in the keymap is too limited
 #this list was generated using:
 #$ DISPLAY=:1 xmodmap -pke | awk -F= '{print $2}' | xargs -n 1 echo | sort -u | grep XF | xargs
-OPTIONAL_KEYS = ["XF86AudioForward", "XF86AudioLowerVolume", "XF86AudioMedia", "XF86AudioMicMute", "XF86AudioMute", "XF86AudioNext", "XF86AudioPause", "XF86AudioPlay", "XF86AudioPrev", "XF86AudioRaiseVolume", "XF86AudioRecord", "XF86AudioRewind", "XF86AudioStop",
-                 "XF86Back", "XF86Battery", "XF86Bluetooth", "XF86Calculator", "XF86ClearGrab", "XF86Close",
-                 "XF86Copy", "XF86Cut", "XF86Display", "XF86Documents", "XF86DOS", "XF86Eject", "XF86Explorer",
-                 "XF86Favorites", "XF86Finance", "XF86Forward", "XF86Game", "XF86Go", "XF86HomePage", "XF86KbdBrightnessDown", "XF86KbdBrightnessUp", "XF86KbdLightOnOff",
-                 "XF86Launch1", "XF86Launch2", "XF86Launch3", "XF86Launch4", "XF86Launch5", "XF86Launch6", "XF86Launch7", "XF86Launch8", "XF86Launch9", "XF86LaunchA", "XF86LaunchB",
-                 "XF86Mail", "XF86MailForward", "XF86MenuKB", "XF86Messenger", "XF86MonBrightnessDown", "XF86MonBrightnessUp", "XF86MyComputer",
-                 "XF86New", "XF86Next_VMode", "XF86Open", "XF86Paste", "XF86Phone", "XF86PowerOff",
-                 "XF86Prev_VMode", "XF86Reload", "XF86Reply", "XF86RotateWindows", "XF86Save", "XF86ScreenSaver",
-                 "XF86ScrollDown", "XF86ScrollUp", "XF86Search", "XF86Send", "XF86Shop", "XF86Sleep", "XF86Suspend",
-                 "XF86Switch_VT_1", "XF86Switch_VT_10", "XF86Switch_VT_11", "XF86Switch_VT_12", "XF86Switch_VT_2", "XF86Switch_VT_3", "XF86Switch_VT_4", "XF86Switch_VT_5", "XF86Switch_VT_6", "XF86Switch_VT_7", "XF86Switch_VT_8", "XF86Switch_VT_9",
-                 "XF86Tools", "XF86TouchpadOff", "XF86TouchpadOn", "XF86TouchpadToggle", "XF86Ungrab", "XF86WakeUp", "XF86WebCam", "XF86WLAN", "XF86WWW", "XF86Xfer"]
+OPTIONAL_KEYS = [
+    "XF86AudioForward", "XF86AudioLowerVolume", "XF86AudioMedia", "XF86AudioMicMute", "XF86AudioMute",
+    "XF86AudioNext", "XF86AudioPause", "XF86AudioPlay", "XF86AudioPrev", "XF86AudioRaiseVolume",
+    "XF86AudioRecord", "XF86AudioRewind", "XF86AudioStop",
+    "XF86Back", "XF86Battery", "XF86Bluetooth", "XF86Calculator", "XF86ClearGrab", "XF86Close",
+    "XF86Copy", "XF86Cut", "XF86Display", "XF86Documents", "XF86DOS", "XF86Eject", "XF86Explorer",
+    "XF86Favorites", "XF86Finance", "XF86Forward", "XF86Game", "XF86Go", "XF86HomePage", "XF86KbdBrightnessDown",
+    "XF86KbdBrightnessUp", "XF86KbdLightOnOff",
+    "XF86Launch1", "XF86Launch2", "XF86Launch3", "XF86Launch4", "XF86Launch5", "XF86Launch6",
+    "XF86Launch7", "XF86Launch8", "XF86Launch9", "XF86LaunchA", "XF86LaunchB",
+    "XF86Mail", "XF86MailForward", "XF86MenuKB", "XF86Messenger", "XF86MonBrightnessDown",
+    "XF86MonBrightnessUp", "XF86MyComputer",
+    "XF86New", "XF86Next_VMode", "XF86Open", "XF86Paste", "XF86Phone", "XF86PowerOff",
+    "XF86Prev_VMode", "XF86Reload", "XF86Reply", "XF86RotateWindows", "XF86Save", "XF86ScreenSaver",
+    "XF86ScrollDown", "XF86ScrollUp", "XF86Search", "XF86Send", "XF86Shop", "XF86Sleep", "XF86Suspend",
+    "XF86Switch_VT_1", "XF86Switch_VT_10", "XF86Switch_VT_11", "XF86Switch_VT_12", "XF86Switch_VT_2", "XF86Switch_VT_3",
+    "XF86Switch_VT_4", "XF86Switch_VT_5", "XF86Switch_VT_6", "XF86Switch_VT_7", "XF86Switch_VT_8", "XF86Switch_VT_9",
+    "XF86Tools", "XF86TouchpadOff", "XF86TouchpadOn", "XF86TouchpadToggle", "XF86Ungrab", "XF86WakeUp", "XF86WebCam",
+    "XF86WLAN", "XF86WWW", "XF86Xfer",
+    ]
 
 
 def clean_keyboard_state():
-    try:
+    with xlog:
         X11Keyboard.ungrab_all_keys()
-    except:
-        log.error("error ungrabbing keys", exc_info=True)
-    try:
+    with xlog:
+        X11Keyboard.set_layout_group(0)
+    with xlog:
         X11Keyboard.unpress_all_keys()
-    except:
-        log.error("error unpressing keys", exc_info=True)
 
 ################################################################################
 # keyboard layouts
@@ -68,20 +77,22 @@ def do_set_keymap(xkbmap_layout, xkbmap_variant, xkbmap_options,
         log("do_set_keymap using xkbmap_query struct=%s", xkbmap_query_struct)
         """ The xkbmap_query_struct data will look something like this:
             {
-            "rules"       : "evdev",
-            "model"       : "pc105",
-            "layout"      : "gb",
-            "options"     : "grp:shift_caps_toggle",
+            b"rules"       : b"evdev",
+            b"model"       : b"pc105",
+            b"layout"      : b"gb",
+            b"options"     : b"grp:shift_caps_toggle",
             }
         """
         #parse the data into a dict:
-        rules = xkbmap_query_struct.get("rules")
-        model = xkbmap_query_struct.get("model")
-        layout = xkbmap_query_struct.get("layout")
-        variant = xkbmap_query_struct.get("variant")
-        options = xkbmap_query_struct.get("options")
+        rules = xkbmap_query_struct.get(b"rules")
+        model = xkbmap_query_struct.get(b"model")
+        layout = xkbmap_query_struct.get(b"layout")
+        variant = xkbmap_query_struct.get(b"variant")
+        options = xkbmap_query_struct.get(b"options")
         if layout:
-            log.info("setting keymap: %s", csv("%s=%s" % (std(k), std(v)) for k,v in xkbmap_query_struct.items() if k in ["rules", "model", "layout"] and v))
+            log.info("setting keymap: %s",
+                     csv("%s=%s" % (std(k), std(v)) for k,v in xkbmap_query_struct.items()
+                         if k in ("rules", "model", "layout") and v))
             safe_setxkbmap(rules, model, layout, variant, options)
         else:
             safe_setxkbmap(rules, model, "", "", "")
@@ -90,7 +101,7 @@ def do_set_keymap(xkbmap_layout, xkbmap_variant, xkbmap_options,
         log("do_set_keymap using xkbmap_print")
         #try to guess the layout by parsing "setxkbmap -print"
         try:
-            sym_re = re.compile("\s*xkb_symbols\s*{\s*include\s*\"([\w\+]*)")
+            sym_re = re.compile(r"\s*xkb_symbols\s*{\s*include\s*\"([\w\+]*)")
             for line in xkbmap_print.splitlines():
                 m = sym_re.match(line)
                 if m:
@@ -110,15 +121,17 @@ def safe_setxkbmap(rules, model, layout, variant, options):
     try:
         X11Keyboard.setxkbmap(rules, model, layout, variant, options)
         return
-    except:
+    except Exception:
+        log("safe_setxkbmap%s", (rules, model, layout, variant, options), exc_info=True)
         log.warn("Warning: failed to set exact keymap,")
-        log.warn(" rules=%s, model=%s, layout=%s, variant=%s, options=%s", rules, model, layout, variant, options)
+        log.warn(" rules=%s, model=%s, layout=%s, variant=%s, options=%s",
+                 bytestostr(rules), bytestostr(model), bytestostr(layout), bytestostr(variant), bytestostr(options))
     if options:
         #try again with no options:
         try:
             X11Keyboard.setxkbmap(rules, model, layout, variant, "")
-            return
-        except:
+        except Exception:
+            log("setxkbmap", exc_info=True)
             log.error("Error: failed to set exact keymap")
             log.error(" even without applying any options..")
 
@@ -130,8 +143,9 @@ def apply_xmodmap(instructions):
     try:
         with xsync:
             unset = X11Keyboard.set_xmodmap(instructions)
-    except:
-        log.error("Error configuring modifier map", exc_info=True)
+    except Exception as e:
+        log("apply_xmodmap(%s)", instructions, exc_info=True)
+        log.error("Error configuring modifier map: %s", e)
         unset = instructions
     if unset is None:
         #None means an X11 error occurred, re-do all:
@@ -139,22 +153,27 @@ def apply_xmodmap(instructions):
     return unset
 
 
+def get_keycode_mappings():
+    if XKB and X11Keyboard.hasXkb():
+        return X11Keyboard.get_xkb_keycode_mappings()
+    return X11Keyboard.get_keycode_mappings()
+
 def set_keycode_translation(xkbmap_x11_keycodes, xkbmap_keycodes):
     """
         Translate the given keycodes into the existing keymap
     """
     #get the list of keycodes (either from x11 keycodes or gtk keycodes):
-    if xkbmap_x11_keycodes and len(xkbmap_x11_keycodes)>0:
+    if xkbmap_x11_keycodes:
         dump_dict(xkbmap_x11_keycodes)
         keycodes = indexed_mappings(xkbmap_x11_keycodes)
     else:
         keycodes = gtk_keycodes_to_mappings(xkbmap_keycodes)
-    log("set_keycode_translation(%s, %s) keycodes=%s", xkbmap_x11_keycodes, xkbmap_keycodes)
+    log("set_keycode_translation(%s, %s)", xkbmap_x11_keycodes, xkbmap_keycodes)
     log(" keycodes=%s", keycodes)
     #keycodes = {
     #     9: set([('', 1), ('Escape', 4), ('', 3), ('Escape', 0), ('Escape', 2)]),
     #     10: set([('onesuperior', 4), ('onesuperior', 8), ('exclam', 1), ('1', 6), ('exclam', 3), ('1', 2), ('exclamdown', 9), ('exclamdown', 5), ('1', 0), ('exclam', 7)]),
-    x11_keycodes = X11Keyboard.get_keycode_mappings()
+    x11_keycodes = get_keycode_mappings()
     log(" x11_keycodes=%s", x11_keycodes)
     #x11_keycodes = {
     #    8: ['Mode_switch', '', 'Mode_switch', '', 'Mode_switch'],
@@ -168,35 +187,41 @@ def set_keycode_translation(xkbmap_x11_keycodes, xkbmap_keycodes):
         keycodes = tuple(x11_keycodes_for_keysym.get(keysym, set()))
         if keysym in DEBUG_KEYSYMS:
             log.info("set_keycode_translation: find_keycode%s x11 keycodes=%s", (kc, keysym, i), keycodes)
-        def rlog(v, msg):
+        def rlog(keycode, msg):
             if keysym in DEBUG_KEYSYMS:
-                log.info("set_keycode_translation: find_keycode%s=%s (%s)", (kc, keysym, i), v, msg)
-            return v
+                log.info("set_keycode_translation: find_keycode%s=%s (%s)", (kc, keysym, i), keycode, msg)
         if not keycodes:
             return None
         #no other option, use it:
-        if len(keycodes)==1:
-            return keycodes[0]
         for keycode in keycodes:
             defs = x11_keycodes.get(keycode)
             if keysym in DEBUG_KEYSYMS:
-                log.info("x11 keycode %i: %s", keycode, defs)
+                log.info("server x11 keycode %i: %s", keycode, defs)
             assert defs, "bug: keycode %i not found in %s" % (keycode, x11_keycodes)
             if len(defs)>i and defs[i]==keysym:
-                return rlog(keycode, "exact index match")
+                rlog(keycode, "exact index match")
+                return keycode, True
         #if possible, use the same one:
         if kc in keycodes:
-            return rlog(kc, "using same keycode as client")
-        return rlog(keycodes[0], "using first match")
+            rlog(kc, "using same keycode as client")
+            return kc, False
+        keycode = keycodes[0]
+        rlog(keycode, "using first match")
+        return keycode, False
     #generate the translation map:
     trans = {}
     for keycode, defs in keycodes.items():
-        for keysym,i in tuple(defs):             #ie: ('1', 0) or ('A', 1), etc
-            x11_keycode = find_keycode(keycode, keysym, i)
-            if x11_keycode:
+        if bool(set(DEBUG_KEYSYMS) & set(bytestostr(d[0]) for d in defs)):
+            log.info("client keycode=%i, defs=%s", keycode, defs)
+        for bkeysym, i in tuple(defs):             #ie: (b'1', 0) or (b'A', 1), etc
+            keysym = bytestostr(bkeysym)
+            m = find_keycode(keycode, keysym, i)
+            if m:
+                x11_keycode, index_matched = m
                 trans[(keycode, keysym)] = x11_keycode
-                trans[(keysym, i)] = x11_keycode
                 trans[keysym] = x11_keycode
+                if index_matched:
+                    trans[(keysym, i)] = x11_keycode
     if not xkbmap_x11_keycodes:
         #now add all the keycodes we may not have mapped yet
         #(present in x11_keycodes but not keycodes)
@@ -229,7 +254,8 @@ def set_all_keycodes(xkbmap_x11_keycodes, xkbmap_keycodes, preserve_server_keyco
         We return a translation map for keycodes after setting them up,
         the key is (keycode, keysym) and the value is the server keycode.
     """
-    log("set_all_keycodes(%s.., %s.., %s.., %s)", str(xkbmap_x11_keycodes)[:60], str(xkbmap_keycodes)[:60], str(preserve_server_keycodes)[:60], modifiers)
+    log("set_all_keycodes(%s.., %s.., %s.., %s)",
+        str(xkbmap_x11_keycodes)[:60], str(xkbmap_keycodes)[:60], str(preserve_server_keycodes)[:60], modifiers)
 
     #so we can validate entries:
     keysym_to_modifier = {}
@@ -263,7 +289,7 @@ def set_all_keycodes(xkbmap_x11_keycodes, xkbmap_keycodes, preserve_server_keyco
         def estr(entries):
             try:
                 return csv(tuple(set(x[0] for x in entries)))
-            except:
+            except Exception:
                 return csv(tuple(entries))
         for keycode, entries in mappings.items():
             mods = modifiers_for(entries)
@@ -280,16 +306,19 @@ def set_all_keycodes(xkbmap_x11_keycodes, xkbmap_keycodes, preserve_server_keyco
                 entries = [entry for entry in entries if mod in modifiers_for([entry])]
                 log.warn(" keeping: %s for %s", estr(entries), mod)
             #now remove entries for keysyms we don't have:
-            f_entries = set([(keysym, index) for keysym, index in entries if keysym and X11Keyboard.parse_keysym(keysym) is not None])
+            f_entries = set((keysym, index) for keysym, index in entries
+                            if keysym and X11Keyboard.parse_keysym(keysym) is not None)
             for keysym, _ in entries:
                 if keysym and X11Keyboard.parse_keysym(keysym) is None:
                     invalid_keysyms.add(keysym)
-            if len(f_entries)==0:
-                log("keymapping removed invalid keycode entry %s pointing to only unknown keysyms: %s", keycode, entries)
+            if not f_entries:
+                log("keymapping removed invalid keycode entry %s pointing to only unknown keysyms: %s",
+                    keycode, entries)
                 continue
             if drop_extra_keys:
-                noopt = [keysym for keysym, index in entries if (X11Keyboard.parse_keysym(keysym) is not None and keysym not in OPTIONAL_KEYS)]
-                if len(noopt)==0:
+                if not any(keysym for keysym, index in entries if (
+                    X11Keyboard.parse_keysym(keysym) is not None and keysym not in OPTIONAL_KEYS)
+                ):
                     log("keymapping removed keycode entry %s pointing to optional keys: %s", keycode, entries)
                     continue
             filtered[keycode] = f_entries
@@ -299,7 +328,7 @@ def set_all_keycodes(xkbmap_x11_keycodes, xkbmap_keycodes, preserve_server_keyco
         return filtered
 
     #get the list of keycodes (either from x11 keycodes or gtk keycodes):
-    if xkbmap_x11_keycodes and len(xkbmap_x11_keycodes)>0:
+    if xkbmap_x11_keycodes:
         log("using x11 keycodes: %s", xkbmap_x11_keycodes)
         dump_dict(xkbmap_x11_keycodes)
         keycodes = indexed_mappings(xkbmap_x11_keycodes)
@@ -322,7 +351,7 @@ def set_all_keycodes(xkbmap_x11_keycodes, xkbmap_keycodes, preserve_server_keyco
         filtered_preserve_keycode_entries = filter_mappings(indexed_mappings(preserve_keycode_entries), try_harder)
         log("filtered_preserve_keycode_entries=%s", filtered_preserve_keycode_entries)
         trans, new_keycodes, missing_keycodes = translate_keycodes(kcmin, kcmax, keycodes, filtered_preserve_keycode_entries, keysym_to_modifier, try_harder)
-        if len(missing_keycodes)==0:
+        if not missing_keycodes:
             break
     instructions = keymap_to_xmodmap(new_keycodes)
     unset = apply_xmodmap(instructions)
@@ -349,8 +378,7 @@ def indexed_mappings(raw_mappings):
     for keycode, keysyms in raw_mappings.items():
         pairs = set()
         l = log
-        for i in range(0, len(keysyms)):
-            keysym = keysyms[i]
+        for i, keysym in enumerate(keysyms):
             if keysym in DEBUG_KEYSYMS:
                 l = log.info
             pairs.add((keysym, i))
@@ -408,7 +436,8 @@ def translate_keycodes(kcmin, kcmax, keycodes, preserve_keycode_entries={}, keys
         Note: a client_keycode of '0' is valid (osx uses that),
         but server_keycode generally starts at 8...
     """
-    log("translate_keycodes(%s, %s, %s, %s, %s, %s)", kcmin, kcmax, keycodes, preserve_keycode_entries, keysym_to_modifier, try_harder)
+    log("translate_keycodes(%s, %s, %s, %s, %s, %s)",
+        kcmin, kcmax, keycodes, preserve_keycode_entries, keysym_to_modifier, try_harder)
     #list of free keycodes we can use:
     free_keycodes = [i for i in range(kcmin, kcmax+1) if i not in preserve_keycode_entries]
     keycode_trans = {}              #translation map from client keycode to our server keycode
@@ -438,7 +467,7 @@ def translate_keycodes(kcmin, kcmax, keycodes, preserve_keycode_entries={}, keys
             l("assign: keycode %s out of range (%s to %s)", server_keycode, kcmin, kcmax)
             server_keycode = -1
         if server_keycode<=0:
-            if len(free_keycodes)>0:
+            if free_keycodes:
                 server_keycode = free_keycodes[0]
                 l("set_keycodes key %s using free keycode=%s", entries, server_keycode)
             else:
@@ -458,18 +487,18 @@ def translate_keycodes(kcmin, kcmax, keycodes, preserve_keycode_entries={}, keys
             for name, _ in entries:
                 if keycode>=0 and server_keycode!=keycode:
                     keycode_trans[(keycode, name)] = server_keycode
-                    l("keycode_trans[(%s, %s)]=%s", keycode, name, server_keycode)
+                    l("keycode_trans[(%s, %s)]=%s", keycode, bytestostr(name), server_keycode)
                 keycode_trans[name] = server_keycode
-                l("keycode_trans[%s]=%s", name, server_keycode)
+                l("keycode_trans[%s]=%s", bytestostr(name), server_keycode)
             server_keycodes[server_keycode] = entries
         return server_keycode
 
     def assign(client_keycode, entries):
-        if len(entries)==0:
+        if not entries:
             return 0
         #all the keysyms for this keycode:
-        keysyms = set([keysym for keysym, _ in entries])
-        if len(keysyms)==0:
+        keysyms = set(keysym for keysym, _ in entries)
+        if not keysyms:
             return 0
         if len(keysyms)==1 and tuple(keysyms)[0]=='0xffffff':
             log("skipped invalid keysym: %s / %s", client_keycode, entries)
@@ -479,7 +508,7 @@ def translate_keycodes(kcmin, kcmax, keycodes, preserve_keycode_entries={}, keys
             l = log.info
         l("assign(%s, %s)", client_keycode, entries)
 
-        if len(preserve_keycode_entries)==0:
+        if not preserve_keycode_entries:
             return do_assign(client_keycode, client_keycode, entries)
         if len(keysyms)==1:
             #only one keysym, replace with single entry
@@ -494,7 +523,7 @@ def translate_keycodes(kcmin, kcmax, keycodes, preserve_keycode_entries={}, keys
                 preserve_keycode_matches[keycode] = v
                 l("preserve_keycode_matches[%s]=%s", keycode, v)
 
-        if len(preserve_keycode_matches)==0:
+        if not preserve_keycode_matches:
             l("no preserve matches for %s", tuple(entries))
             return do_assign(client_keycode, -1, entries)         #nothing to preserve
 
@@ -502,15 +531,17 @@ def translate_keycodes(kcmin, kcmax, keycodes, preserve_keycode_entries={}, keys
         #direct superset:
         for p_keycode, p_entries in preserve_keycode_matches.items():
             if entries.issubset(p_entries):
-                l("found direct preserve superset for client keycode %s %s -> server keycode %s %s", client_keycode, tuple(entries), p_keycode, tuple(p_entries))
+                l("found direct preserve superset for client keycode %s %s -> server keycode %s %s",
+                  client_keycode, tuple(entries), p_keycode, tuple(p_entries))
                 return do_assign(client_keycode, p_keycode, p_entries, override_server_keycode=True)
             if p_entries.issubset(entries):
-                l("found direct superset of preserve for client keycode %s %s -> server keycode %s %s", client_keycode, tuple(entries), p_keycode, tuple(p_entries))
+                l("found direct superset of preserve for client keycode %s %s -> server keycode %s %s",
+                  client_keycode, tuple(entries), p_keycode, tuple(p_entries))
                 return do_assign(client_keycode, p_keycode, entries, override_server_keycode=True)
 
         #ignoring indexes, but requiring at least as many keysyms:
         for p_keycode, p_entries in preserve_keycode_matches.items():
-            p_keysyms = set([keysym for keysym,_ in p_entries])
+            p_keysyms = set(keysym for keysym,_ in p_entries)
             if keysyms.issubset(p_keysyms):
                 if len(p_entries)>len(entries):
                     l("found keysym preserve superset with more keys for %s : %s", tuple(entries), tuple(p_entries))
@@ -521,10 +552,10 @@ def translate_keycodes(kcmin, kcmax, keycodes, preserve_keycode_entries={}, keys
 
         if try_harder:
             #try to match the main key only:
-            main_key = set([(keysym, index) for keysym, index in entries if index==0])
+            main_key = set((keysym, index) for keysym, index in entries if index==0)
             if len(main_key)==1:
                 for p_keycode, p_entries in preserve_keycode_matches.items():
-                    p_keysyms = set([keysym for keysym,_ in p_entries])
+                    p_keysyms = set(keysym for keysym,_ in p_entries)
                     if main_key.issubset(p_entries):
                         l("found main key superset for %s : %s", main_key, tuple(p_entries))
                         return do_assign(client_keycode, p_keycode, p_entries, override_server_keycode=True)
@@ -585,10 +616,10 @@ def keymap_to_xmodmap(trans_keycodes):
         names = [""]*keysyms_per_keycode
         sentries = sorted(entries, key=lambda x:x[1])
         for name, index in sentries:
-            assert 0<=index and index<keysyms_per_keycode
+            assert 0<=index<keysyms_per_keycode
             try:
                 keysym = X11Keyboard.parse_keysym(name)
-            except:
+            except Exception:
                 keysym = None
             if keysym is None:
                 if name!="":
@@ -620,7 +651,7 @@ def keymap_to_xmodmap(trans_keycodes):
                 if name in DEBUG_KEYSYMS:
                     log.info("keymap_to_xmodmap: keysyms[%s]=%s (%s)", index, keysym, name)
         #remove empty keysyms:
-        while len(keysyms)>0 and keysyms[0] is None:
+        while keysyms and keysyms[0] is None:
             keysyms = keysyms[1:]
         l = log
         if [k for k in keysyms if k in DEBUG_KEYSYMS]:
@@ -628,7 +659,7 @@ def keymap_to_xmodmap(trans_keycodes):
         l("%s: %s -> %s", server_keycode, names, keysyms)
         instructions.append(("keycode", server_keycode, keysyms))
 
-    if len(missing_keysyms)>0:
+    if missing_keysyms:
         log.error("cannot find the X11 keysym for the following key names: %s", set(missing_keysyms))
     log("instructions=%s", instructions)
     return  instructions
@@ -650,16 +681,16 @@ def set_modifiers(modifiers):
     """
     instructions = []
     for modifier, keynames in modifiers.items():
-        mod = X11Keyboard.parse_modifier(modifier)
+        mod = X11Keyboard.parse_modifier(bytestostr(modifier))
         if mod>=0:
             instructions.append(("add", mod, keynames))
         else:
-            log.error("set_modifiers_from_dict: unknown modifier %s", modifier)
+            log.error("Error: unknown modifier %s", modifier)
     log("set_modifiers: %s", instructions)
     def apply_or_trim(instructions):
         err = apply_xmodmap(instructions)
         log("set_modifiers: err=%s", err)
-        if len(err):
+        if err:
             log("set_modifiers %s failed, retrying one more at a time", instructions)
             l = len(instructions)
             for i in range(1, l):
@@ -667,7 +698,7 @@ def set_modifiers(modifiers):
                 log("set_modifiers testing with [:%s]=%s", i, subset)
                 err = apply_xmodmap(subset)
                 log("err=%s", err)
-                if len(err)>0:
+                if err:
                     log.warn("removing problematic modifier mapping: %s", instructions[i-1])
                     instructions = instructions[:i-1]+instructions[i:]
                     return apply_or_trim(instructions)
@@ -683,9 +714,10 @@ def get_modifiers_from_meanings(xkbmap_mod_meanings):
     #first generate a {modifier : [keynames]} dict:
     modifiers = {}
     for keyname, modifier in xkbmap_mod_meanings.items():
-        l = modifiers.setdefault(modifier, [])
-        if keyname not in l:
-            l.append(keyname)
+        l = modifiers.setdefault(bytestostr(modifier), [])
+        kn = bytestostr(keyname)
+        if kn not in l:
+            l.append(kn)
     log("get_modifiers_from_meanings(%s) modifier dict=%s", xkbmap_mod_meanings, modifiers)
     return modifiers
 
@@ -704,8 +736,9 @@ def get_modifiers_from_keycodes(xkbmap_keycodes, add_default_modifiers=True):
         _, keyname, _, _, _ = entry
         modifier = pref.get(keyname)
         if modifier:
-            keynames = matches.setdefault(modifier, set())
-            keynames.add(keyname)
+            keynames = matches.setdefault(modifier, [])
+            if keyname not in keynames:
+                keynames.add(keyname)
             all_keynames.add(keyname)
     if add_default_modifiers:
         #try to add missings ones (magic!)
@@ -715,10 +748,14 @@ def get_modifiers_from_keycodes(xkbmap_keycodes, add_default_modifiers=True):
                 continue            #aleady defined
             if modifier not in matches:
                 #define it since it is completely missing
-                defaults.setdefault(modifier, set()).add(keyname)
+                keynames = defaults.setdefault(modifier, [])
+                if keyname not in keynames:
+                    keynames.append(keyname)
             elif modifier in ["shift", "lock", "control", "mod1", "mod2"] or keyname=="ISO_Level3_Shift":
                 #these ones we always add them, even if a record for this modifier already exists
-                matches.setdefault(modifier, set()).add(keyname)
+                keynames = matches.setdefault(modifier, [])
+                if keyname not in keynames:
+                    keynames.append(keyname)
         log("get_modifiers_from_keycodes(...) adding defaults: %s", defaults)
         matches.update(defaults)
     log("get_modifiers_from_keycodes(...)=%s", matches)
@@ -728,7 +765,8 @@ def map_missing_modifiers(keynames_for_mod):
     x11_keycodes = X11Keyboard.get_keycode_mappings()
     min_keycode, max_keycode = X11Keyboard.get_minmax_keycodes()
     free_keycodes = [x for x in range(min_keycode, max_keycode) if x not in x11_keycodes]
-    log("map_missing_modifiers(%s) min_keycode=%i max_keycode=%i, free_keycodes=%s", keynames_for_mod, min_keycode, max_keycode, free_keycodes)
+    log("map_missing_modifiers(%s) min_keycode=%i max_keycode=%i, free_keycodes=%s",
+        keynames_for_mod, min_keycode, max_keycode, free_keycodes)
     keysyms_to_keycode = {}
     for keycode, keysyms in x11_keycodes.items():
         for keysym in keysyms:
